package java;

import java.io.*;
import java.util.*;

public class Cluster {

  public void clear() {
    for (int i = 0; i < m_size; i++) {
      m_vertex_used[i] = false;
      m_clusters.clear();
    }

  }

  public Cluster(String distance_file)
      throws IOException {
    m_size = -1;
    m_clusters = new ArrayList(5);
    BufferedReader file = new BufferedReader(new FileReader(distance_file));
    String line = file.readLine();
    String splits[] = line.split(" ");
    m_size = splits.length;
    m_distance = new double[m_size][m_size];
    m_vertex_used = new boolean[m_size];
    populateDistMatrix(line, 0);
    int counter = 0;
    while ((line = file.readLine()) != null) {
      counter++;
      populateDistMatrix(line, counter);
    }
    file.close();
    if (counter != m_size - 1) {
      throw new RuntimeException((new StringBuilder()).append("The number of roes and columns do not match: ").append(counter).append("  ").append(m_size).toString());
    }
    m_vertex_used = new boolean[m_size];
    for (int i = 0; i < m_vertex_used.length; i++) { m_vertex_used[i] = false; }

  }

  private void populateDistMatrix(String line, int row_index) {
    String splits[] = line.split(" ");
    for (int i = 0; i < splits.length; i++) {
      double val = Double.parseDouble(splits[i]);
      m_distance[row_index][i] = val;
    }

  }

  public void cluster() {
    for (int left_Points = m_size; left_Points > 0; ) {
      int next_point = getNextUnusedPointIndex();
      if (RANDOMNESS_ON) {
        double random = (new Random(System.currentTimeMillis())).nextInt(3);
        if (random == 1.0D) {
          while (random != 1.0D) {
            random = (new Random(System.currentTimeMillis())).nextInt(200);
            next_point = getNextUnusedPointIndex();
          }
        } else { next_point = getNextunusedPointBasedOnSizeHeuristic(); }
      }
      int total_cluster_distance = 0;
      ArrayList new_cluster = new ArrayList(m_size / 5);
      int i;
      for (i = 0; i < m_size; i++) {
        if (next_point != i && !m_vertex_used[i] && m_distance[next_point][i] <= 0.5D) {
          total_cluster_distance = (int) ((double) total_cluster_distance + m_distance[next_point][i]);
          new_cluster.add(Integer.valueOf(i + 1));
        }
      }

      if ((double) total_cluster_distance / (double) new_cluster.size() >= 0.25D) { new_cluster.clear(); }
      new_cluster.add(Integer.valueOf(next_point + 1));
      Collections.sort(new_cluster);
      m_clusters.add(new_cluster);
      left_Points -= new_cluster.size();
      i = 0;
      while (i < new_cluster.size()) {
        m_vertex_used[((Integer) new_cluster.get(i)).intValue() - 1] = true;
        i++;
      }
    }

  }

  private int getNextUnusedPointIndex() {
    for (int i = 0; i < m_vertex_used.length; i++) { if (!m_vertex_used[i]) { return i; } }

    return -1;
  }

  private int getNextunusedPointBasedOnSizeHeuristic() {
    int smallestClusterSize = 0x7fffffff;
    int smallestClsuterCenter = -1;
    int biggestClusterSize = 0;
    int biggestClusterCenter = -1;
    HashMap clusters = new HashMap(10);
    for (int point = 0; point < m_vertex_used.length; point++) {
      if (m_vertex_used[point]) { continue; }
      int total_cluster_distance = 0;
      ArrayList new_cluster = new ArrayList(m_size / 5);
      for (int i = 0; i < m_size; i++) {
        if (point != i && !m_vertex_used[i] && m_distance[point][i] <= 0.5D) {
          total_cluster_distance = (int) ((double) total_cluster_distance + m_distance[point][i]);
          new_cluster.add(Integer.valueOf(i + 1));
        }
      }

      if ((double) total_cluster_distance / (double) new_cluster.size() >= 0.25D) { new_cluster.clear(); }
      new_cluster.add(Integer.valueOf(point + 1));
      if (new_cluster.size() < smallestClusterSize) {
        smallestClsuterCenter = point;
        smallestClusterSize = new_cluster.size();
      }
      if (new_cluster.size() > biggestClusterSize) {
        biggestClusterSize = new_cluster.size();
        biggestClusterCenter = point;
      }
    }

    if (SmallestClusterFirst) { return smallestClsuterCenter; } else { return biggestClusterCenter; }
  }

  public void dumpClusters2File(String filename)
      throws IOException {
    BufferedWriter file = new BufferedWriter(new FileWriter(filename));
    for (int i = 0; i < m_clusters.size(); i++) {
      ArrayList cluster = (ArrayList) m_clusters.get(i);
      for (int j = 0; j < cluster.size(); j++) { file.write((new StringBuilder()).append(cluster.get(j).toString()).append("\t").toString()); }

      file.write("\n");
    }

    file.flush();
    file.close();
  }

  public static void main(String args[])
      throws IOException {
    if (args.length < 3) {
      System.out.println("Usage: Cluster <dist mat file> <adj_mat file> <op cluster file>");
      return;
    }
    Cluster clOb = new Cluster(args[0]);
    double best_mod = 0.0D;
    int best_sample = -1;
    int num_tries = 1;
    for (int i = 0; i < num_tries; i++) {
      clOb.cluster();
      clOb.dumpClusters2File((new StringBuilder()).append("temp.sample_").append(i).toString());
      clOb.clear();
      MiscTools miscTollObject = new MiscTools();
      miscTollObject.caclModularity(args[1], (new StringBuilder()).append("temp.sample_").append(i).toString());
      if (miscTollObject.getModularity() > best_mod) {
        best_mod = miscTollObject.getModularity();
        best_sample = i;
      }
    }

    System.out.println((new StringBuilder()).append("Best Mod:").append(best_mod).toString());
    File bestFile = new File((new StringBuilder()).append("temp.sample_").append(best_sample).toString());
    File destinationFile = new File(args[2]);
    bestFile.renameTo(destinationFile);
    for (int i = 0; i < num_tries; i++) {
      if (i != best_sample) {
        (new File((new StringBuilder()).append("temp.sample_").append(i).toString())).delete();
      }
    }

  }

  private static boolean RANDOMNESS_ON = false;
  private static boolean SmallestClusterFirst;
  private static boolean BiggestClusterFirst;
  private double m_distance[][];
  private boolean m_vertex_used[];
  private int m_size;
  private static final double BoundaryLimit = 0.5D;
  private static final double AvgLimit = 0.25D;
  private ArrayList m_clusters;

  static {
    SmallestClusterFirst = false;
    BiggestClusterFirst = !SmallestClusterFirst;
  }
}
