package mod;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
import java.text.DecimalFormat;
import java.text.ParseException;

/**
 * User: Gaurav Agarwal
 * Date: May 6, 2007
 * Time: 8:56:06 PM
 * <p/>
 * Class to represent the dendogram.
 */
public class DendoGram
{

    private double m_max_mod = 0;
    private double m_currentMod = 0;
    private int m_maxModSeqId = -1;
    private int m_freeId = 0;


    private HashMap<Integer, Node> m_id2nodeMap = new HashMap<Integer, Node>();
    private ArrayList<Integer> m_splitSequence = new ArrayList<Integer>();
    private ArrayList<Double> m_modChangeSequence = new ArrayList<Double>();


    private TreeSet<Node> m_heap = new TreeSet<Node>(new Comparator()
    {
        public int compare(Object o1, Object o2)
        {
            Node ob1 = (Node) o1;
            Node ob2 = (Node) o2;

            if (ob1.getId() == ob2.getId())
            {
                return 0;
            }
            else if (ob1.getDelta() < ob2.getDelta())
            {
                return 1;
            }
            else
            {
                return -1;
            }
        }
    });

    public void addRoot(int[] partition, int[] left, int[] right, double delta)
    {
        Node parent = new Node(partition, m_freeId++);
        parent.setDelta(delta);

        Node child1Node = new Node(left, m_freeId++);
        Node child2Node = new Node(right, m_freeId++);
        parent.setLeft(child1Node);
        parent.setRight(child2Node);

        m_heap.add(parent);
        m_id2nodeMap.put(parent.getId(), parent);
        m_id2nodeMap.put(child1Node.getId(), child1Node);
        m_id2nodeMap.put(child2Node.getId(), child2Node);


        System.out.println("Called addRoot");
        System.out.println("Partition Size: " + partition.length);
        System.out.println("left size: " + left.length);
        System.out.println("right size: " + right.length);
        System.out.println("delta: " + delta);
        System.out.println("Heap size:\n" + m_heap.size());

    }

    public void addPartition(int parentId, int[] child1, int[] child2, double delta)
    {
        System.out.println("Called addPartition");
        System.out.println("parentId:" + parentId);
        Node parent = m_id2nodeMap.get(parentId);
        parent.setDelta(delta);
        if (child1 != null)
        {
            Node child1Node = new Node(child1, m_freeId++);
            parent.setLeft(child1Node);
            m_id2nodeMap.put(child1Node.getId(), child1Node);
            System.out.println("child1 size: " + child1.length);
        }
        if (child2 != null)

        {
            Node child2Node = new Node(child2, m_freeId++);
            parent.setRight(child2Node);
            m_id2nodeMap.put(child2Node.getId(), child2Node);
            System.out.println("child2 size: " + child2.length);
        }
        m_heap.add(parent);
        System.out.println("Delta: " + delta);
        System.out.println("Heap size: \n" + m_heap.size());
    }

    public Node[] splitForMaxDelta()
    {
        System.out.println("Called splitForMax");
        Node parent = m_heap.first();
        if (parent.getDelta() <= 0)
        {
            System.out.println("No split changing delta for better.. I'm putting an end to this.");
            m_heap.clear();
            return new Node[2];
        }

        m_heap.remove(parent);
        /* while (parent.getDelta() == 0)
        {
            if (m_heap.size() == 0)
            {
                System.out.println("No split changing delta for better.. I'm putting an end to this.");
                return new Node[2];
            }
            parent = m_heap.first();
            m_heap.remove(parent);
        }*/


        System.out.println("Splitting node with id: " + parent.getId());
        System.out.print("Node Elements: ");
        printArray(parent.getPoints());
        System.out.println("");


        Node[] children = new Node[2];
        children[0] = parent.getLeft();
        children[1] = parent.getRight();

        System.out.print("Left Elements: ");
        if (children[0] != null)
        {
            printArray(children[0].getPoints());
        }
        System.out.println("");

        System.out.print("Right Elements: ");
        if (children[1] != null)
        {
            printArray(children[1].getPoints());
        }
        System.out.println("");

        m_currentMod += parent.getDelta();

        if (m_currentMod > m_max_mod)
        {
            m_max_mod = m_currentMod;
            m_maxModSeqId = parent.getId();
        }
        m_splitSequence.add(parent.getId());
        m_modChangeSequence.add(m_currentMod);

        System.out.print("Mod Change Sequence: ");
        for (int i = 0; i < m_modChangeSequence.size(); i++)
        {
            System.out.print(m_modChangeSequence.get(i) + "\t");
        }
        System.out.println("");

        System.out.print("Split Sequence: ");
        for (int i = 0; i < m_splitSequence.size(); i++)
        {
            System.out.print(m_splitSequence.get(i) + "\t");
        }
        System.out.println("\n");
        return children;
    }

    private void printArray(int[] points)
    {
        for (int i = 0; i < points.length; i++)
        {
            System.out.print(" " + points[i]);
        }
    }

    public int elemtntsLeftForSplitting()
    {
        return m_heap.size();
    }

    public void dumpcluster2File(String fileName) throws IOException
    {
        BufferedWriter out = new BufferedWriter(new FileWriter(fileName));
        HashSet<Node> clusters = new HashSet<Node>();
        System.out.println("Dump2File Called");
        System.out.println("MaxModSeq:" + m_maxModSeqId);
        System.out.println("MaxMod: " + m_max_mod);
        if (m_maxModSeqId > -1)
        {
            for (int i = 0; i < m_splitSequence.size(); i++)
            {
                int splitNodeId = m_splitSequence.get(i);
                Node splitNode = m_id2nodeMap.get(splitNodeId);
                clusters.remove(splitNode);
                clusters.add(splitNode.getLeft());
                clusters.add(splitNode.getRight());
                if (m_maxModSeqId == splitNodeId)
                {
                    break;
                }
            }
        }

        int it = 0;
        for (Node node : clusters)
        {
            int[] points = node.getPoints();
            System.out.print("Cluster " + ++it + " :");
            for (int point : points)
            {
                System.out.print(point + "\t");
                out.write(Integer.toString(point) + "\t");
            }
            out.write("\n");
            System.out.println("");
        }
        out.close();
    }
}

